#
# Copyright (C) 2003 Rational Discovery LLC
#    All Rights Reserved
#

availableCanvases = {
  'pdf': ('PDF.pidPDF', 'PDFCanvas', 'PDF'),
  'ps': ('PS.pidPS', 'PSCanvas', 'PS'),
  'svg': ('SVG.pidSVG', 'SVGCanvas', 'SVG'),
  'jpg': ('PIL.pidPIL', 'PILCanvas', 'JPEG'),
  'png': ('PIL.pidPIL', 'PILCanvas', 'PNG')
}
