// auto-generated export definition header
#pragma once
#include <RDGeneral/RDExportMacros.h>

// RDKIT_ABBREVIATIONS_EXPORT definitions
#ifdef RDKIT_ABBREVIATIONS_LOCAL_BUILD
#define RDKIT_ABBREVIATIONS_EXPORT
#elif defined(RDKIT_ABBREVIATIONS_BUILD)
#define RDKIT_ABBREVIATIONS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_ABBREVIATIONS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_ABBREVIATIONS_EXPORT end definitions

// RDKIT_ALIGNMENT_EXPORT definitions
#ifdef RDKIT_ALIGNMENT_LOCAL_BUILD
#define RDKIT_ALIGNMENT_EXPORT
#elif defined(RDKIT_ALIGNMENT_BUILD)
#define RDKIT_ALIGNMENT_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_ALIGNMENT_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_ALIGNMENT_EXPORT end definitions

// RDKIT_AVALONLIB_EXPORT definitions
#ifdef RDKIT_AVALONLIB_LOCAL_BUILD
#define RDKIT_AVALONLIB_EXPORT
#elif defined(RDKIT_AVALONLIB_BUILD)
#define RDKIT_AVALONLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_AVALONLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_AVALONLIB_EXPORT end definitions

// RDKIT_CIPLABELER_EXPORT definitions
#ifdef RDKIT_CIPLABELER_LOCAL_BUILD
#define RDKIT_CIPLABELER_EXPORT
#elif defined(RDKIT_CIPLABELER_BUILD)
#define RDKIT_CIPLABELER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CIPLABELER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CIPLABELER_EXPORT end definitions

// RDKIT_CATALOGS_EXPORT definitions
#ifdef RDKIT_CATALOGS_LOCAL_BUILD
#define RDKIT_CATALOGS_EXPORT
#elif defined(RDKIT_CATALOGS_BUILD)
#define RDKIT_CATALOGS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CATALOGS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CATALOGS_EXPORT end definitions

// RDKIT_CHEMDRAW_EXPORT definitions
#ifdef RDKIT_CHEMDRAW_LOCAL_BUILD
#define RDKIT_CHEMDRAW_EXPORT
#elif defined(RDKIT_CHEMDRAW_BUILD)
#define RDKIT_CHEMDRAW_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CHEMDRAW_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CHEMDRAW_EXPORT end definitions

// RDKIT_CHEMREACTIONS_EXPORT definitions
#ifdef RDKIT_CHEMREACTIONS_LOCAL_BUILD
#define RDKIT_CHEMREACTIONS_EXPORT
#elif defined(RDKIT_CHEMREACTIONS_BUILD)
#define RDKIT_CHEMREACTIONS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CHEMREACTIONS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CHEMREACTIONS_EXPORT end definitions

// RDKIT_CHEMTRANSFORMS_EXPORT definitions
#ifdef RDKIT_CHEMTRANSFORMS_LOCAL_BUILD
#define RDKIT_CHEMTRANSFORMS_EXPORT
#elif defined(RDKIT_CHEMTRANSFORMS_BUILD)
#define RDKIT_CHEMTRANSFORMS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CHEMTRANSFORMS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CHEMTRANSFORMS_EXPORT end definitions

// RDKIT_CHEMICALFEATURES_EXPORT definitions
#ifdef RDKIT_CHEMICALFEATURES_LOCAL_BUILD
#define RDKIT_CHEMICALFEATURES_EXPORT
#elif defined(RDKIT_CHEMICALFEATURES_BUILD)
#define RDKIT_CHEMICALFEATURES_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CHEMICALFEATURES_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CHEMICALFEATURES_EXPORT end definitions

// RDKIT_CONFORMERPARSER_EXPORT definitions
#ifdef RDKIT_CONFORMERPARSER_LOCAL_BUILD
#define RDKIT_CONFORMERPARSER_EXPORT
#elif defined(RDKIT_CONFORMERPARSER_BUILD)
#define RDKIT_CONFORMERPARSER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_CONFORMERPARSER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_CONFORMERPARSER_EXPORT end definitions

// RDKIT_DATASTRUCTS_EXPORT definitions
#ifdef RDKIT_DATASTRUCTS_LOCAL_BUILD
#define RDKIT_DATASTRUCTS_EXPORT
#elif defined(RDKIT_DATASTRUCTS_BUILD)
#define RDKIT_DATASTRUCTS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DATASTRUCTS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DATASTRUCTS_EXPORT end definitions

// RDKIT_DEPICTOR_EXPORT definitions
#ifdef RDKIT_DEPICTOR_LOCAL_BUILD
#define RDKIT_DEPICTOR_EXPORT
#elif defined(RDKIT_DEPICTOR_BUILD)
#define RDKIT_DEPICTOR_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DEPICTOR_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DEPICTOR_EXPORT end definitions

// RDKIT_DEPROTECT_EXPORT definitions
#ifdef RDKIT_DEPROTECT_LOCAL_BUILD
#define RDKIT_DEPROTECT_EXPORT
#elif defined(RDKIT_DEPROTECT_BUILD)
#define RDKIT_DEPROTECT_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DEPROTECT_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DEPROTECT_EXPORT end definitions

// RDKIT_DESCRIPTORS_EXPORT definitions
#ifdef RDKIT_DESCRIPTORS_LOCAL_BUILD
#define RDKIT_DESCRIPTORS_EXPORT
#elif defined(RDKIT_DESCRIPTORS_BUILD)
#define RDKIT_DESCRIPTORS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DESCRIPTORS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DESCRIPTORS_EXPORT end definitions

// RDKIT_DETERMINEBONDS_EXPORT definitions
#ifdef RDKIT_DETERMINEBONDS_LOCAL_BUILD
#define RDKIT_DETERMINEBONDS_EXPORT
#elif defined(RDKIT_DETERMINEBONDS_BUILD)
#define RDKIT_DETERMINEBONDS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DETERMINEBONDS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DETERMINEBONDS_EXPORT end definitions

// RDKIT_DISTGEOMHELPERS_EXPORT definitions
#ifdef RDKIT_DISTGEOMHELPERS_LOCAL_BUILD
#define RDKIT_DISTGEOMHELPERS_EXPORT
#elif defined(RDKIT_DISTGEOMHELPERS_BUILD)
#define RDKIT_DISTGEOMHELPERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DISTGEOMHELPERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DISTGEOMHELPERS_EXPORT end definitions

// RDKIT_DISTGEOMETRY_EXPORT definitions
#ifdef RDKIT_DISTGEOMETRY_LOCAL_BUILD
#define RDKIT_DISTGEOMETRY_EXPORT
#elif defined(RDKIT_DISTGEOMETRY_BUILD)
#define RDKIT_DISTGEOMETRY_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_DISTGEOMETRY_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_DISTGEOMETRY_EXPORT end definitions

// RDKIT_EHTLIB_EXPORT definitions
#ifdef RDKIT_EHTLIB_LOCAL_BUILD
#define RDKIT_EHTLIB_EXPORT
#elif defined(RDKIT_EHTLIB_BUILD)
#define RDKIT_EHTLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_EHTLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_EHTLIB_EXPORT end definitions

// RDKIT_EIGENSOLVERS_EXPORT definitions
#ifdef RDKIT_EIGENSOLVERS_LOCAL_BUILD
#define RDKIT_EIGENSOLVERS_EXPORT
#elif defined(RDKIT_EIGENSOLVERS_BUILD)
#define RDKIT_EIGENSOLVERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_EIGENSOLVERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_EIGENSOLVERS_EXPORT end definitions

// RDKIT_ENUMERATESTEREOISOMERS_EXPORT definitions
#ifdef RDKIT_ENUMERATESTEREOISOMERS_LOCAL_BUILD
#define RDKIT_ENUMERATESTEREOISOMERS_EXPORT
#elif defined(RDKIT_ENUMERATESTEREOISOMERS_BUILD)
#define RDKIT_ENUMERATESTEREOISOMERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_ENUMERATESTEREOISOMERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_ENUMERATESTEREOISOMERS_EXPORT end definitions

// RDKIT_FMCS_EXPORT definitions
#ifdef RDKIT_FMCS_LOCAL_BUILD
#define RDKIT_FMCS_EXPORT
#elif defined(RDKIT_FMCS_BUILD)
#define RDKIT_FMCS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FMCS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FMCS_EXPORT end definitions

// RDKIT_FILEPARSERS_EXPORT definitions
#ifdef RDKIT_FILEPARSERS_LOCAL_BUILD
#define RDKIT_FILEPARSERS_EXPORT
#elif defined(RDKIT_FILEPARSERS_BUILD)
#define RDKIT_FILEPARSERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FILEPARSERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FILEPARSERS_EXPORT end definitions

// RDKIT_FILTERCATALOG_EXPORT definitions
#ifdef RDKIT_FILTERCATALOG_LOCAL_BUILD
#define RDKIT_FILTERCATALOG_EXPORT
#elif defined(RDKIT_FILTERCATALOG_BUILD)
#define RDKIT_FILTERCATALOG_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FILTERCATALOG_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FILTERCATALOG_EXPORT end definitions

// RDKIT_FINGERPRINTS_EXPORT definitions
#ifdef RDKIT_FINGERPRINTS_LOCAL_BUILD
#define RDKIT_FINGERPRINTS_EXPORT
#elif defined(RDKIT_FINGERPRINTS_BUILD)
#define RDKIT_FINGERPRINTS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FINGERPRINTS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FINGERPRINTS_EXPORT end definitions

// RDKIT_FORCEFIELD_EXPORT definitions
#ifdef RDKIT_FORCEFIELD_LOCAL_BUILD
#define RDKIT_FORCEFIELD_EXPORT
#elif defined(RDKIT_FORCEFIELD_BUILD)
#define RDKIT_FORCEFIELD_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FORCEFIELD_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FORCEFIELD_EXPORT end definitions

// RDKIT_FORCEFIELDHELPERS_EXPORT definitions
#ifdef RDKIT_FORCEFIELDHELPERS_LOCAL_BUILD
#define RDKIT_FORCEFIELDHELPERS_EXPORT
#elif defined(RDKIT_FORCEFIELDHELPERS_BUILD)
#define RDKIT_FORCEFIELDHELPERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FORCEFIELDHELPERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FORCEFIELDHELPERS_EXPORT end definitions

// RDKIT_FRAGCATALOG_EXPORT definitions
#ifdef RDKIT_FRAGCATALOG_LOCAL_BUILD
#define RDKIT_FRAGCATALOG_EXPORT
#elif defined(RDKIT_FRAGCATALOG_BUILD)
#define RDKIT_FRAGCATALOG_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FRAGCATALOG_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FRAGCATALOG_EXPORT end definitions

// RDKIT_FREESASALIB_EXPORT definitions
#ifdef RDKIT_FREESASALIB_LOCAL_BUILD
#define RDKIT_FREESASALIB_EXPORT
#elif defined(RDKIT_FREESASALIB_BUILD)
#define RDKIT_FREESASALIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FREESASALIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FREESASALIB_EXPORT end definitions

// RDKIT_GAUSSIANSHAPE_EXPORT definitions
#ifdef RDKIT_GAUSSIANSHAPE_LOCAL_BUILD
#define RDKIT_GAUSSIANSHAPE_EXPORT
#elif defined(RDKIT_GAUSSIANSHAPE_BUILD)
#define RDKIT_GAUSSIANSHAPE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_GAUSSIANSHAPE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_GAUSSIANSHAPE_EXPORT end definitions

// RDKIT_GENERALIZEDSUBSTRUCT_EXPORT definitions
#ifdef RDKIT_GENERALIZEDSUBSTRUCT_LOCAL_BUILD
#define RDKIT_GENERALIZEDSUBSTRUCT_EXPORT
#elif defined(RDKIT_GENERALIZEDSUBSTRUCT_BUILD)
#define RDKIT_GENERALIZEDSUBSTRUCT_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_GENERALIZEDSUBSTRUCT_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_GENERALIZEDSUBSTRUCT_EXPORT end definitions

// RDKIT_GENERICGROUPS_EXPORT definitions
#ifdef RDKIT_GENERICGROUPS_LOCAL_BUILD
#define RDKIT_GENERICGROUPS_EXPORT
#elif defined(RDKIT_GENERICGROUPS_BUILD)
#define RDKIT_GENERICGROUPS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_GENERICGROUPS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_GENERICGROUPS_EXPORT end definitions

// RDKIT_GRAPHMOL_EXPORT definitions
#ifdef RDKIT_GRAPHMOL_LOCAL_BUILD
#define RDKIT_GRAPHMOL_EXPORT
#elif defined(RDKIT_GRAPHMOL_BUILD)
#define RDKIT_GRAPHMOL_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_GRAPHMOL_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_GRAPHMOL_EXPORT end definitions

// RDKIT_INCHI_EXPORT definitions
#ifdef RDKIT_INCHI_LOCAL_BUILD
#define RDKIT_INCHI_EXPORT
#elif defined(RDKIT_INCHI_BUILD)
#define RDKIT_INCHI_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_INCHI_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_INCHI_EXPORT end definitions

// RDKIT_INFOTHEORY_EXPORT definitions
#ifdef RDKIT_INFOTHEORY_LOCAL_BUILD
#define RDKIT_INFOTHEORY_EXPORT
#elif defined(RDKIT_INFOTHEORY_BUILD)
#define RDKIT_INFOTHEORY_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_INFOTHEORY_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_INFOTHEORY_EXPORT end definitions

// RDKIT_MMPA_EXPORT definitions
#ifdef RDKIT_MMPA_LOCAL_BUILD
#define RDKIT_MMPA_EXPORT
#elif defined(RDKIT_MMPA_BUILD)
#define RDKIT_MMPA_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MMPA_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MMPA_EXPORT end definitions

// RDKIT_MARVINPARSER_EXPORT definitions
#ifdef RDKIT_MARVINPARSER_LOCAL_BUILD
#define RDKIT_MARVINPARSER_EXPORT
#elif defined(RDKIT_MARVINPARSER_BUILD)
#define RDKIT_MARVINPARSER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MARVINPARSER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MARVINPARSER_EXPORT end definitions

// RDKIT_MOLALIGN_EXPORT definitions
#ifdef RDKIT_MOLALIGN_LOCAL_BUILD
#define RDKIT_MOLALIGN_EXPORT
#elif defined(RDKIT_MOLALIGN_BUILD)
#define RDKIT_MOLALIGN_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLALIGN_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLALIGN_EXPORT end definitions

// RDKIT_MOLCATALOG_EXPORT definitions
#ifdef RDKIT_MOLCATALOG_LOCAL_BUILD
#define RDKIT_MOLCATALOG_EXPORT
#elif defined(RDKIT_MOLCATALOG_BUILD)
#define RDKIT_MOLCATALOG_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLCATALOG_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLCATALOG_EXPORT end definitions

// RDKIT_MOLCHEMICALFEATURES_EXPORT definitions
#ifdef RDKIT_MOLCHEMICALFEATURES_LOCAL_BUILD
#define RDKIT_MOLCHEMICALFEATURES_EXPORT
#elif defined(RDKIT_MOLCHEMICALFEATURES_BUILD)
#define RDKIT_MOLCHEMICALFEATURES_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLCHEMICALFEATURES_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLCHEMICALFEATURES_EXPORT end definitions

// RDKIT_MOLDRAW2D_EXPORT definitions
#ifdef RDKIT_MOLDRAW2D_LOCAL_BUILD
#define RDKIT_MOLDRAW2D_EXPORT
#elif defined(RDKIT_MOLDRAW2D_BUILD)
#define RDKIT_MOLDRAW2D_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLDRAW2D_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLDRAW2D_EXPORT end definitions

// RDKIT_MOLDRAW2DQT_EXPORT definitions
#ifdef RDKIT_MOLDRAW2DQT_LOCAL_BUILD
#define RDKIT_MOLDRAW2DQT_EXPORT
#elif defined(RDKIT_MOLDRAW2DQT_BUILD)
#define RDKIT_MOLDRAW2DQT_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLDRAW2DQT_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLDRAW2DQT_EXPORT end definitions

// RDKIT_MOLENUMERATOR_EXPORT definitions
#ifdef RDKIT_MOLENUMERATOR_LOCAL_BUILD
#define RDKIT_MOLENUMERATOR_EXPORT
#elif defined(RDKIT_MOLENUMERATOR_BUILD)
#define RDKIT_MOLENUMERATOR_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLENUMERATOR_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLENUMERATOR_EXPORT end definitions

// RDKIT_MOLHASH_EXPORT definitions
#ifdef RDKIT_MOLHASH_LOCAL_BUILD
#define RDKIT_MOLHASH_EXPORT
#elif defined(RDKIT_MOLHASH_BUILD)
#define RDKIT_MOLHASH_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLHASH_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLHASH_EXPORT end definitions

// RDKIT_MOLINTERACTIONFIELDS_EXPORT definitions
#ifdef RDKIT_MOLINTERACTIONFIELDS_LOCAL_BUILD
#define RDKIT_MOLINTERACTIONFIELDS_EXPORT
#elif defined(RDKIT_MOLINTERACTIONFIELDS_BUILD)
#define RDKIT_MOLINTERACTIONFIELDS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLINTERACTIONFIELDS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLINTERACTIONFIELDS_EXPORT end definitions

// RDKIT_MOLINTERCHANGE_EXPORT definitions
#ifdef RDKIT_MOLINTERCHANGE_LOCAL_BUILD
#define RDKIT_MOLINTERCHANGE_EXPORT
#elif defined(RDKIT_MOLINTERCHANGE_BUILD)
#define RDKIT_MOLINTERCHANGE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLINTERCHANGE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLINTERCHANGE_EXPORT end definitions

// RDKIT_MOLPROCESSING_EXPORT definitions
#ifdef RDKIT_MOLPROCESSING_LOCAL_BUILD
#define RDKIT_MOLPROCESSING_EXPORT
#elif defined(RDKIT_MOLPROCESSING_BUILD)
#define RDKIT_MOLPROCESSING_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLPROCESSING_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLPROCESSING_EXPORT end definitions

// RDKIT_MOLSTANDARDIZE_EXPORT definitions
#ifdef RDKIT_MOLSTANDARDIZE_LOCAL_BUILD
#define RDKIT_MOLSTANDARDIZE_EXPORT
#elif defined(RDKIT_MOLSTANDARDIZE_BUILD)
#define RDKIT_MOLSTANDARDIZE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLSTANDARDIZE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLSTANDARDIZE_EXPORT end definitions

// RDKIT_MOLTRANSFORMS_EXPORT definitions
#ifdef RDKIT_MOLTRANSFORMS_LOCAL_BUILD
#define RDKIT_MOLTRANSFORMS_EXPORT
#elif defined(RDKIT_MOLTRANSFORMS_BUILD)
#define RDKIT_MOLTRANSFORMS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MOLTRANSFORMS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MOLTRANSFORMS_EXPORT end definitions

// RDKIT_O3AALIGN_EXPORT definitions
#ifdef RDKIT_O3AALIGN_LOCAL_BUILD
#define RDKIT_O3AALIGN_EXPORT
#elif defined(RDKIT_O3AALIGN_BUILD)
#define RDKIT_O3AALIGN_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_O3AALIGN_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_O3AALIGN_EXPORT end definitions

// RDKIT_OPTIMIZER_EXPORT definitions
#ifdef RDKIT_OPTIMIZER_LOCAL_BUILD
#define RDKIT_OPTIMIZER_EXPORT
#elif defined(RDKIT_OPTIMIZER_BUILD)
#define RDKIT_OPTIMIZER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_OPTIMIZER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_OPTIMIZER_EXPORT end definitions

// RDKIT_PARTIALCHARGES_EXPORT definitions
#ifdef RDKIT_PARTIALCHARGES_LOCAL_BUILD
#define RDKIT_PARTIALCHARGES_EXPORT
#elif defined(RDKIT_PARTIALCHARGES_BUILD)
#define RDKIT_PARTIALCHARGES_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_PARTIALCHARGES_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_PARTIALCHARGES_EXPORT end definitions

// RDKIT_PUBCHEMSHAPE_EXPORT definitions
#ifdef RDKIT_PUBCHEMSHAPE_LOCAL_BUILD
#define RDKIT_PUBCHEMSHAPE_EXPORT
#elif defined(RDKIT_PUBCHEMSHAPE_BUILD)
#define RDKIT_PUBCHEMSHAPE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_PUBCHEMSHAPE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_PUBCHEMSHAPE_EXPORT end definitions

// RDKIT_RDBOOST_EXPORT definitions
#ifdef RDKIT_RDBOOST_LOCAL_BUILD
#define RDKIT_RDBOOST_EXPORT
#elif defined(RDKIT_RDBOOST_BUILD)
#define RDKIT_RDBOOST_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDBOOST_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDBOOST_EXPORT end definitions

// RDKIT_RDCHEMDRAWLIB_EXPORT definitions
#ifdef RDKIT_RDCHEMDRAWLIB_LOCAL_BUILD
#define RDKIT_RDCHEMDRAWLIB_EXPORT
#elif defined(RDKIT_RDCHEMDRAWLIB_BUILD)
#define RDKIT_RDCHEMDRAWLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDCHEMDRAWLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDCHEMDRAWLIB_EXPORT end definitions

// RDKIT_RDCHEMDRAWREACTIONLIB_EXPORT definitions
#ifdef RDKIT_RDCHEMDRAWREACTIONLIB_LOCAL_BUILD
#define RDKIT_RDCHEMDRAWREACTIONLIB_EXPORT
#elif defined(RDKIT_RDCHEMDRAWREACTIONLIB_BUILD)
#define RDKIT_RDCHEMDRAWREACTIONLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDCHEMDRAWREACTIONLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDCHEMDRAWREACTIONLIB_EXPORT end definitions

// RDKIT_RDGENERAL_EXPORT definitions
#ifdef RDKIT_RDGENERAL_LOCAL_BUILD
#define RDKIT_RDGENERAL_EXPORT
#elif defined(RDKIT_RDGENERAL_BUILD)
#define RDKIT_RDGENERAL_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDGENERAL_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDGENERAL_EXPORT end definitions

// RDKIT_RDGEOMETRYLIB_EXPORT definitions
#ifdef RDKIT_RDGEOMETRYLIB_LOCAL_BUILD
#define RDKIT_RDGEOMETRYLIB_EXPORT
#elif defined(RDKIT_RDGEOMETRYLIB_BUILD)
#define RDKIT_RDGEOMETRYLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDGEOMETRYLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDGEOMETRYLIB_EXPORT end definitions

// RDKIT_RDINCHILIB_EXPORT definitions
#ifdef RDKIT_RDINCHILIB_LOCAL_BUILD
#define RDKIT_RDINCHILIB_EXPORT
#elif defined(RDKIT_RDINCHILIB_BUILD)
#define RDKIT_RDINCHILIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDINCHILIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDINCHILIB_EXPORT end definitions

// RDKIT_RDSTREAMS_EXPORT definitions
#ifdef RDKIT_RDSTREAMS_LOCAL_BUILD
#define RDKIT_RDSTREAMS_EXPORT
#elif defined(RDKIT_RDSTREAMS_BUILD)
#define RDKIT_RDSTREAMS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RDSTREAMS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RDSTREAMS_EXPORT end definitions

// RDKIT_RGROUPDECOMPOSITION_EXPORT definitions
#ifdef RDKIT_RGROUPDECOMPOSITION_LOCAL_BUILD
#define RDKIT_RGROUPDECOMPOSITION_EXPORT
#elif defined(RDKIT_RGROUPDECOMPOSITION_BUILD)
#define RDKIT_RGROUPDECOMPOSITION_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RGROUPDECOMPOSITION_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RGROUPDECOMPOSITION_EXPORT end definitions

// RDKIT_RASCALMCES_EXPORT definitions
#ifdef RDKIT_RASCALMCES_LOCAL_BUILD
#define RDKIT_RASCALMCES_EXPORT
#elif defined(RDKIT_RASCALMCES_BUILD)
#define RDKIT_RASCALMCES_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RASCALMCES_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RASCALMCES_EXPORT end definitions

// RDKIT_REDUCEDGRAPHS_EXPORT definitions
#ifdef RDKIT_REDUCEDGRAPHS_LOCAL_BUILD
#define RDKIT_REDUCEDGRAPHS_EXPORT
#elif defined(RDKIT_REDUCEDGRAPHS_BUILD)
#define RDKIT_REDUCEDGRAPHS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_REDUCEDGRAPHS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_REDUCEDGRAPHS_EXPORT end definitions

// RDKIT_RINGDECOMPOSERLIB_EXPORT definitions
#ifdef RDKIT_RINGDECOMPOSERLIB_LOCAL_BUILD
#define RDKIT_RINGDECOMPOSERLIB_EXPORT
#elif defined(RDKIT_RINGDECOMPOSERLIB_BUILD)
#define RDKIT_RINGDECOMPOSERLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_RINGDECOMPOSERLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_RINGDECOMPOSERLIB_EXPORT end definitions

// RDKIT_SLNPARSE_EXPORT definitions
#ifdef RDKIT_SLNPARSE_LOCAL_BUILD
#define RDKIT_SLNPARSE_EXPORT
#elif defined(RDKIT_SLNPARSE_BUILD)
#define RDKIT_SLNPARSE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SLNPARSE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SLNPARSE_EXPORT end definitions

// RDKIT_SCAFFOLDNETWORK_EXPORT definitions
#ifdef RDKIT_SCAFFOLDNETWORK_LOCAL_BUILD
#define RDKIT_SCAFFOLDNETWORK_EXPORT
#elif defined(RDKIT_SCAFFOLDNETWORK_BUILD)
#define RDKIT_SCAFFOLDNETWORK_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SCAFFOLDNETWORK_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SCAFFOLDNETWORK_EXPORT end definitions

// RDKIT_SHAPEHELPERS_EXPORT definitions
#ifdef RDKIT_SHAPEHELPERS_LOCAL_BUILD
#define RDKIT_SHAPEHELPERS_EXPORT
#elif defined(RDKIT_SHAPEHELPERS_BUILD)
#define RDKIT_SHAPEHELPERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SHAPEHELPERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SHAPEHELPERS_EXPORT end definitions

// RDKIT_SIMDIVPICKERS_EXPORT definitions
#ifdef RDKIT_SIMDIVPICKERS_LOCAL_BUILD
#define RDKIT_SIMDIVPICKERS_EXPORT
#elif defined(RDKIT_SIMDIVPICKERS_BUILD)
#define RDKIT_SIMDIVPICKERS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SIMDIVPICKERS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SIMDIVPICKERS_EXPORT end definitions

// RDKIT_SMILESPARSE_EXPORT definitions
#ifdef RDKIT_SMILESPARSE_LOCAL_BUILD
#define RDKIT_SMILESPARSE_EXPORT
#elif defined(RDKIT_SMILESPARSE_BUILD)
#define RDKIT_SMILESPARSE_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SMILESPARSE_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SMILESPARSE_EXPORT end definitions

// RDKIT_STRUCTCHECKER_EXPORT definitions
#ifdef RDKIT_STRUCTCHECKER_LOCAL_BUILD
#define RDKIT_STRUCTCHECKER_EXPORT
#elif defined(RDKIT_STRUCTCHECKER_BUILD)
#define RDKIT_STRUCTCHECKER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_STRUCTCHECKER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_STRUCTCHECKER_EXPORT end definitions

// RDKIT_SUBGRAPHS_EXPORT definitions
#ifdef RDKIT_SUBGRAPHS_LOCAL_BUILD
#define RDKIT_SUBGRAPHS_EXPORT
#elif defined(RDKIT_SUBGRAPHS_BUILD)
#define RDKIT_SUBGRAPHS_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SUBGRAPHS_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SUBGRAPHS_EXPORT end definitions

// RDKIT_SUBSTRUCTLIBRARY_EXPORT definitions
#ifdef RDKIT_SUBSTRUCTLIBRARY_LOCAL_BUILD
#define RDKIT_SUBSTRUCTLIBRARY_EXPORT
#elif defined(RDKIT_SUBSTRUCTLIBRARY_BUILD)
#define RDKIT_SUBSTRUCTLIBRARY_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SUBSTRUCTLIBRARY_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SUBSTRUCTLIBRARY_EXPORT end definitions

// RDKIT_SUBSTRUCTMATCH_EXPORT definitions
#ifdef RDKIT_SUBSTRUCTMATCH_LOCAL_BUILD
#define RDKIT_SUBSTRUCTMATCH_EXPORT
#elif defined(RDKIT_SUBSTRUCTMATCH_BUILD)
#define RDKIT_SUBSTRUCTMATCH_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SUBSTRUCTMATCH_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SUBSTRUCTMATCH_EXPORT end definitions

// RDKIT_SYNTHONSPACESEARCH_EXPORT definitions
#ifdef RDKIT_SYNTHONSPACESEARCH_LOCAL_BUILD
#define RDKIT_SYNTHONSPACESEARCH_EXPORT
#elif defined(RDKIT_SYNTHONSPACESEARCH_BUILD)
#define RDKIT_SYNTHONSPACESEARCH_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_SYNTHONSPACESEARCH_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_SYNTHONSPACESEARCH_EXPORT end definitions

// RDKIT_TAUTOMERQUERY_EXPORT definitions
#ifdef RDKIT_TAUTOMERQUERY_LOCAL_BUILD
#define RDKIT_TAUTOMERQUERY_EXPORT
#elif defined(RDKIT_TAUTOMERQUERY_BUILD)
#define RDKIT_TAUTOMERQUERY_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_TAUTOMERQUERY_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_TAUTOMERQUERY_EXPORT end definitions

// RDKIT_TRAJECTORY_EXPORT definitions
#ifdef RDKIT_TRAJECTORY_LOCAL_BUILD
#define RDKIT_TRAJECTORY_EXPORT
#elif defined(RDKIT_TRAJECTORY_BUILD)
#define RDKIT_TRAJECTORY_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_TRAJECTORY_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_TRAJECTORY_EXPORT end definitions

// RDKIT_AVALON_CLIB_EXPORT definitions
#ifdef RDKIT_AVALON_CLIB_LOCAL_BUILD
#define RDKIT_AVALON_CLIB_EXPORT
#elif defined(RDKIT_AVALON_CLIB_BUILD)
#define RDKIT_AVALON_CLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_AVALON_CLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_AVALON_CLIB_EXPORT end definitions

// RDKIT_COORDGEN_EXPORT definitions
#ifdef RDKIT_COORDGEN_LOCAL_BUILD
#define RDKIT_COORDGEN_EXPORT
#elif defined(RDKIT_COORDGEN_BUILD)
#define RDKIT_COORDGEN_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_COORDGEN_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_COORDGEN_EXPORT end definitions

// RDKIT_FREESASA_CLIB_EXPORT definitions
#ifdef RDKIT_FREESASA_CLIB_LOCAL_BUILD
#define RDKIT_FREESASA_CLIB_EXPORT
#elif defined(RDKIT_FREESASA_CLIB_BUILD)
#define RDKIT_FREESASA_CLIB_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_FREESASA_CLIB_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_FREESASA_CLIB_EXPORT end definitions

// RDKIT_GA_EXPORT definitions
#ifdef RDKIT_GA_LOCAL_BUILD
#define RDKIT_GA_EXPORT
#elif defined(RDKIT_GA_BUILD)
#define RDKIT_GA_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_GA_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_GA_EXPORT end definitions

// RDKIT_HC_EXPORT definitions
#ifdef RDKIT_HC_LOCAL_BUILD
#define RDKIT_HC_EXPORT
#elif defined(RDKIT_HC_BUILD)
#define RDKIT_HC_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_HC_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_HC_EXPORT end definitions

// RDKIT_MAEPARSER_EXPORT definitions
#ifdef RDKIT_MAEPARSER_LOCAL_BUILD
#define RDKIT_MAEPARSER_EXPORT
#elif defined(RDKIT_MAEPARSER_BUILD)
#define RDKIT_MAEPARSER_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_MAEPARSER_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_MAEPARSER_EXPORT end definitions

// RDKIT_PUBCHEM_ALIGN3D_EXPORT definitions
#ifdef RDKIT_PUBCHEM_ALIGN3D_LOCAL_BUILD
#define RDKIT_PUBCHEM_ALIGN3D_EXPORT
#elif defined(RDKIT_PUBCHEM_ALIGN3D_BUILD)
#define RDKIT_PUBCHEM_ALIGN3D_EXPORT RDKIT_EXPORT_API
#else
#define RDKIT_PUBCHEM_ALIGN3D_EXPORT RDKIT_IMPORT_API
#endif
// RDKIT_PUBCHEM_ALIGN3D_EXPORT end definitions

/*
 * Do not dll export/import to export queries (it will mess up with the
 * templates), but make sure it is visible for *nix
 */
// RDKIT_QUERY_EXPORT definitions
#if defined(RDKIT_DYN_LINK) && defined(WIN32) && defined(BOOST_HAS_DECLSPEC)
#define RDKIT_QUERY_EXPORT
#else
#define RDKIT_QUERY_EXPORT RDKIT_GRAPHMOL_EXPORT
#endif
// RDKIT_QUERY_EXPORT end definitions
