// auto-generated header to be imported in all cpp tests
#pragma once

#ifdef RDKIT_ABBREVIATIONS_BUILD
#undef RDKIT_ABBREVIATIONS_BUILD
#endif

#ifdef RDKIT_ALIGNMENT_BUILD
#undef RDKIT_ALIGNMENT_BUILD
#endif

#ifdef RDKIT_AVALONLIB_BUILD
#undef RDKIT_AVALONLIB_BUILD
#endif

#ifdef RDKIT_CIPLABELER_BUILD
#undef RDKIT_CIPLABELER_BUILD
#endif

#ifdef RDKIT_CATALOGS_BUILD
#undef RDKIT_CATALOGS_BUILD
#endif

#ifdef RDKIT_CHEMDRAW_BUILD
#undef RDKIT_CHEMDRAW_BUILD
#endif

#ifdef RDKIT_CHEMREACTIONS_BUILD
#undef RDKIT_CHEMREACTIONS_BUILD
#endif

#ifdef RDKIT_CHEMTRANSFORMS_BUILD
#undef RDKIT_CHEMTRANSFORMS_BUILD
#endif

#ifdef RDKIT_CHEMICALFEATURES_BUILD
#undef RDKIT_CHEMICALFEATURES_BUILD
#endif

#ifdef RDKIT_CONFORMERPARSER_BUILD
#undef RDKIT_CONFORMERPARSER_BUILD
#endif

#ifdef RDKIT_DATASTRUCTS_BUILD
#undef RDKIT_DATASTRUCTS_BUILD
#endif

#ifdef RDKIT_DEPICTOR_BUILD
#undef RDKIT_DEPICTOR_BUILD
#endif

#ifdef RDKIT_DEPROTECT_BUILD
#undef RDKIT_DEPROTECT_BUILD
#endif

#ifdef RDKIT_DESCRIPTORS_BUILD
#undef RDKIT_DESCRIPTORS_BUILD
#endif

#ifdef RDKIT_DETERMINEBONDS_BUILD
#undef RDKIT_DETERMINEBONDS_BUILD
#endif

#ifdef RDKIT_DISTGEOMHELPERS_BUILD
#undef RDKIT_DISTGEOMHELPERS_BUILD
#endif

#ifdef RDKIT_DISTGEOMETRY_BUILD
#undef RDKIT_DISTGEOMETRY_BUILD
#endif

#ifdef RDKIT_EHTLIB_BUILD
#undef RDKIT_EHTLIB_BUILD
#endif

#ifdef RDKIT_EIGENSOLVERS_BUILD
#undef RDKIT_EIGENSOLVERS_BUILD
#endif

#ifdef RDKIT_ENUMERATESTEREOISOMERS_BUILD
#undef RDKIT_ENUMERATESTEREOISOMERS_BUILD
#endif

#ifdef RDKIT_FMCS_BUILD
#undef RDKIT_FMCS_BUILD
#endif

#ifdef RDKIT_FILEPARSERS_BUILD
#undef RDKIT_FILEPARSERS_BUILD
#endif

#ifdef RDKIT_FILTERCATALOG_BUILD
#undef RDKIT_FILTERCATALOG_BUILD
#endif

#ifdef RDKIT_FINGERPRINTS_BUILD
#undef RDKIT_FINGERPRINTS_BUILD
#endif

#ifdef RDKIT_FORCEFIELD_BUILD
#undef RDKIT_FORCEFIELD_BUILD
#endif

#ifdef RDKIT_FORCEFIELDHELPERS_BUILD
#undef RDKIT_FORCEFIELDHELPERS_BUILD
#endif

#ifdef RDKIT_FRAGCATALOG_BUILD
#undef RDKIT_FRAGCATALOG_BUILD
#endif

#ifdef RDKIT_FREESASALIB_BUILD
#undef RDKIT_FREESASALIB_BUILD
#endif

#ifdef RDKIT_GAUSSIANSHAPE_BUILD
#undef RDKIT_GAUSSIANSHAPE_BUILD
#endif

#ifdef RDKIT_GENERALIZEDSUBSTRUCT_BUILD
#undef RDKIT_GENERALIZEDSUBSTRUCT_BUILD
#endif

#ifdef RDKIT_GENERICGROUPS_BUILD
#undef RDKIT_GENERICGROUPS_BUILD
#endif

#ifdef RDKIT_GRAPHMOL_BUILD
#undef RDKIT_GRAPHMOL_BUILD
#endif

#ifdef RDKIT_INCHI_BUILD
#undef RDKIT_INCHI_BUILD
#endif

#ifdef RDKIT_INFOTHEORY_BUILD
#undef RDKIT_INFOTHEORY_BUILD
#endif

#ifdef RDKIT_MMPA_BUILD
#undef RDKIT_MMPA_BUILD
#endif

#ifdef RDKIT_MARVINPARSER_BUILD
#undef RDKIT_MARVINPARSER_BUILD
#endif

#ifdef RDKIT_MOLALIGN_BUILD
#undef RDKIT_MOLALIGN_BUILD
#endif

#ifdef RDKIT_MOLCATALOG_BUILD
#undef RDKIT_MOLCATALOG_BUILD
#endif

#ifdef RDKIT_MOLCHEMICALFEATURES_BUILD
#undef RDKIT_MOLCHEMICALFEATURES_BUILD
#endif

#ifdef RDKIT_MOLDRAW2D_BUILD
#undef RDKIT_MOLDRAW2D_BUILD
#endif

#ifdef RDKIT_MOLDRAW2DQT_BUILD
#undef RDKIT_MOLDRAW2DQT_BUILD
#endif

#ifdef RDKIT_MOLENUMERATOR_BUILD
#undef RDKIT_MOLENUMERATOR_BUILD
#endif

#ifdef RDKIT_MOLHASH_BUILD
#undef RDKIT_MOLHASH_BUILD
#endif

#ifdef RDKIT_MOLINTERACTIONFIELDS_BUILD
#undef RDKIT_MOLINTERACTIONFIELDS_BUILD
#endif

#ifdef RDKIT_MOLINTERCHANGE_BUILD
#undef RDKIT_MOLINTERCHANGE_BUILD
#endif

#ifdef RDKIT_MOLPROCESSING_BUILD
#undef RDKIT_MOLPROCESSING_BUILD
#endif

#ifdef RDKIT_MOLSTANDARDIZE_BUILD
#undef RDKIT_MOLSTANDARDIZE_BUILD
#endif

#ifdef RDKIT_MOLTRANSFORMS_BUILD
#undef RDKIT_MOLTRANSFORMS_BUILD
#endif

#ifdef RDKIT_O3AALIGN_BUILD
#undef RDKIT_O3AALIGN_BUILD
#endif

#ifdef RDKIT_OPTIMIZER_BUILD
#undef RDKIT_OPTIMIZER_BUILD
#endif

#ifdef RDKIT_PARTIALCHARGES_BUILD
#undef RDKIT_PARTIALCHARGES_BUILD
#endif

#ifdef RDKIT_PUBCHEMSHAPE_BUILD
#undef RDKIT_PUBCHEMSHAPE_BUILD
#endif

#ifdef RDKIT_RDBOOST_BUILD
#undef RDKIT_RDBOOST_BUILD
#endif

#ifdef RDKIT_RDCHEMDRAWLIB_BUILD
#undef RDKIT_RDCHEMDRAWLIB_BUILD
#endif

#ifdef RDKIT_RDCHEMDRAWREACTIONLIB_BUILD
#undef RDKIT_RDCHEMDRAWREACTIONLIB_BUILD
#endif

#ifdef RDKIT_RDGENERAL_BUILD
#undef RDKIT_RDGENERAL_BUILD
#endif

#ifdef RDKIT_RDGEOMETRYLIB_BUILD
#undef RDKIT_RDGEOMETRYLIB_BUILD
#endif

#ifdef RDKIT_RDINCHILIB_BUILD
#undef RDKIT_RDINCHILIB_BUILD
#endif

#ifdef RDKIT_RDSTREAMS_BUILD
#undef RDKIT_RDSTREAMS_BUILD
#endif

#ifdef RDKIT_RGROUPDECOMPOSITION_BUILD
#undef RDKIT_RGROUPDECOMPOSITION_BUILD
#endif

#ifdef RDKIT_RASCALMCES_BUILD
#undef RDKIT_RASCALMCES_BUILD
#endif

#ifdef RDKIT_REDUCEDGRAPHS_BUILD
#undef RDKIT_REDUCEDGRAPHS_BUILD
#endif

#ifdef RDKIT_RINGDECOMPOSERLIB_BUILD
#undef RDKIT_RINGDECOMPOSERLIB_BUILD
#endif

#ifdef RDKIT_SLNPARSE_BUILD
#undef RDKIT_SLNPARSE_BUILD
#endif

#ifdef RDKIT_SCAFFOLDNETWORK_BUILD
#undef RDKIT_SCAFFOLDNETWORK_BUILD
#endif

#ifdef RDKIT_SHAPEHELPERS_BUILD
#undef RDKIT_SHAPEHELPERS_BUILD
#endif

#ifdef RDKIT_SIMDIVPICKERS_BUILD
#undef RDKIT_SIMDIVPICKERS_BUILD
#endif

#ifdef RDKIT_SMILESPARSE_BUILD
#undef RDKIT_SMILESPARSE_BUILD
#endif

#ifdef RDKIT_STRUCTCHECKER_BUILD
#undef RDKIT_STRUCTCHECKER_BUILD
#endif

#ifdef RDKIT_SUBGRAPHS_BUILD
#undef RDKIT_SUBGRAPHS_BUILD
#endif

#ifdef RDKIT_SUBSTRUCTLIBRARY_BUILD
#undef RDKIT_SUBSTRUCTLIBRARY_BUILD
#endif

#ifdef RDKIT_SUBSTRUCTMATCH_BUILD
#undef RDKIT_SUBSTRUCTMATCH_BUILD
#endif

#ifdef RDKIT_SYNTHONSPACESEARCH_BUILD
#undef RDKIT_SYNTHONSPACESEARCH_BUILD
#endif

#ifdef RDKIT_TAUTOMERQUERY_BUILD
#undef RDKIT_TAUTOMERQUERY_BUILD
#endif

#ifdef RDKIT_TRAJECTORY_BUILD
#undef RDKIT_TRAJECTORY_BUILD
#endif

#ifdef RDKIT_AVALON_CLIB_BUILD
#undef RDKIT_AVALON_CLIB_BUILD
#endif

#ifdef RDKIT_COORDGEN_BUILD
#undef RDKIT_COORDGEN_BUILD
#endif

#ifdef RDKIT_FREESASA_CLIB_BUILD
#undef RDKIT_FREESASA_CLIB_BUILD
#endif

#ifdef RDKIT_GA_BUILD
#undef RDKIT_GA_BUILD
#endif

#ifdef RDKIT_HC_BUILD
#undef RDKIT_HC_BUILD
#endif

#ifdef RDKIT_MAEPARSER_BUILD
#undef RDKIT_MAEPARSER_BUILD
#endif

#ifdef RDKIT_PUBCHEM_ALIGN3D_BUILD
#undef RDKIT_PUBCHEM_ALIGN3D_BUILD
#endif
