//
//  Copyright (C) 2002-2006 Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//

/*! \file RDKitQueries.h

  \brief pulls in the \c RDKit Query functionality

*/
#include <RDGeneral/export.h>
#ifndef _RDKIT_QUERIES_H
#define _RDKIT_QUERIES_H

#include <Query/QueryObjects.h>
#include <GraphMol/QueryAtom.h>
#include <GraphMol/QueryBond.h>
#include <GraphMol/QueryOps.h>

#endif
