//
//  Copyright (C) 2013 Paolo Tosco
//
//  Copyright (C) 2004-2006 Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include <RDGeneral/export.h>
#ifndef __RD_MMFFCONTRIBS_H__
#define __RD_MMFFCONTRIBS_H__
/*! \file Contribs.h
  \brief This convenience include pulls in all the other MMFF include files
*/

#include "BondStretch.h"
#include "AngleBend.h"
#include "StretchBend.h"
#include "OopBend.h"
#include "TorsionAngle.h"
#include "Nonbonded.h"

#endif
