//
//  Copyright (C) 2001-2006 Greg Landrum and Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
/*! \file GraphMol.h

  \brief pulls in RWMol and ROMol

*/
#include <RDGeneral/export.h>
#ifndef _GRAPH_MOL_H
#define _GRAPH_MOL_H

#include "ROMol.h"
#include "RWMol.h"

#endif
