//
//  Copyright (C) 2016 Greg Landrum
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//

#include <RDGeneral/export.h>
#ifndef RD_MOLDESCRIPTORS3D_H_SEPT2016
#define RD_MOLDESCRIPTORS3D_H_SEPT2016

#include <GraphMol/Descriptors/CoulombMat.h>
#include <GraphMol/Descriptors/EEM.h>
#include <GraphMol/Descriptors/PBF.h>
#include <GraphMol/Descriptors/RDF.h>
#include <GraphMol/Descriptors/MORSE.h>
#include <GraphMol/Descriptors/WHIM.h>
#include <GraphMol/Descriptors/GETAWAY.h>
#include <GraphMol/Descriptors/AUTOCORR3D.h>
#include <GraphMol/Descriptors/PMI.h>

#endif
