//
// Copyright (c) 2001-2006 greg Landrum and Rational Discovery LLC
//
//  @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
/*! \file BitVects.h

  \brief Pulls in all the BitVect classes

*/
#include <RDGeneral/export.h>
#ifndef __BITVECTS_H__
#define __BITVECTS_H__

#include "BitVect.h"
#include "ExplicitBitVect.h"
#include "SparseBitVect.h"
typedef SparseBitVect SBV;
typedef ExplicitBitVect EBV;

#endif
